from __future__ import print_function
import json
import boto3
import os
import mimetypes
from zipfile import ZipFile
from io import BytesIO
import logging
from .env_utils import is_stg_or_prd_env, prod_bucket_env
from .misc_utils import PRINT


###########################
# Config
###########################
logging.basicConfig()
logger = logging.getLogger(__name__)


class s3Utils(object):  # NOQA - This class name violates style rules, but a lot of things might break if we change it.

    SYS_BUCKET_TEMPLATE = "elasticbeanstalk-%s-system"
    OUTFILE_BUCKET_TEMPLATE = "elasticbeanstalk-%s-wfoutput"
    RAW_BUCKET_TEMPLATE = "elasticbeanstalk-%s-files"
    BLOB_BUCKET_TEMPLATE = "elasticbeanstalk-%s-blobs"

    def __init__(self, outfile_bucket=None, sys_bucket=None, raw_file_bucket=None,
                 blob_bucket=None, env=None):
        """
        if we pass in env set the outfile and sys bucket from the environment
        """
        # avoid circular ref
        from .beanstalk_utils import get_beanstalk_real_url
        self.url = ''
        self.s3 = boto3.client('s3', region_name='us-east-1')
        if sys_bucket is None:
            # staging and production share same buckets
            if env:
                if is_stg_or_prd_env(env):
                    self.url = get_beanstalk_real_url(env)
                    env = prod_bucket_env(env)
            # we use standardized naming schema, so s3 buckets always have same prefix
            sys_bucket = self.SYS_BUCKET_TEMPLATE % env
            outfile_bucket = self.OUTFILE_BUCKET_TEMPLATE % env
            raw_file_bucket = self.RAW_BUCKET_TEMPLATE % env
            blob_bucket = self.BLOB_BUCKET_TEMPLATE % env

        self.sys_bucket = sys_bucket
        self.outfile_bucket = outfile_bucket
        self.raw_file_bucket = raw_file_bucket
        self.blob_bucket = blob_bucket

    ACCESS_KEYS_S3_KEY = 'access_key_admin'

    def get_access_keys(self, name=ACCESS_KEYS_S3_KEY):
        keys = self.get_key(keyfile_name=name)
        if not isinstance(keys, dict):
            raise ValueError("Remotely stored access keys are not in the expected form")

        if isinstance(keys.get('default'), dict):
            keys = keys['default']
        if self.url:
            keys['server'] = self.url
        return keys

    def get_ff_key(self):
        return self.get_access_keys()

    def get_higlass_key(self):
        # higlass key corresponds to Django server super user credentials
        return self.get_key(keyfile_name='api_key_higlass')

    def get_google_key(self):
        return self.get_key(keyfile_name='api_key_google')

    def get_jupyterhub_key(self):
        # jupyterhub key is a Jupyterhub API token
        return self.get_key(keyfile_name='api_key_jupyterhub')

    def get_key(self, keyfile_name='access_key_admin'):
        # Share secret encrypted S3 File
        response = self.s3.get_object(Bucket=self.sys_bucket,
                                      Key=keyfile_name,
                                      SSECustomerKey=os.environ['S3_ENCRYPT_KEY'],
                                      SSECustomerAlgorithm='AES256')
        akey = response['Body'].read()
        if type(akey) == bytes:
            akey = akey.decode()
        try:
            return json.loads(akey)
        except (ValueError, TypeError):
            # maybe its not json after all
            return akey

    def read_s3(self, filename):
        response = self.s3.get_object(Bucket=self.outfile_bucket, Key=filename)
        logger.info(str(response))
        return response['Body'].read()

    def does_key_exist(self, key, bucket=None, print_error=True):
        if not bucket:
            bucket = self.outfile_bucket
        try:
            file_metadata = self.s3.head_object(Bucket=bucket, Key=key)
        except Exception as e:
            if print_error:
                PRINT("object %s not found on bucket %s" % (str(key), str(bucket)))
                PRINT(str(e))
            return False
        return file_metadata

    def get_file_size(self, key, bucket=None, add_bytes=0, add_gb=0,
                      size_in_gb=False):
        """
        default returns file size in bytes,
        unless size_in_gb = True
        """
        meta = self.does_key_exist(key, bucket)
        if not meta:
            raise Exception("key not found")
        one_gb = 1073741824
        add = add_bytes + (add_gb * one_gb)
        size = meta['ContentLength'] + add
        if size_in_gb:
            size = size / one_gb
        return size

    def delete_key(self, key, bucket=None):
        if not bucket:
            bucket = self.outfile_bucket
        self.s3.delete_object(Bucket=bucket, Key=key)

    @classmethod
    def size(cls, bucket):
        sbuck = boto3.resource('s3').Bucket(bucket)
        # get only head of objects so we can count them
        return sum(1 for _ in sbuck.objects.all())

    def s3_put(self, obj, upload_key, acl=None):
        """
        try to guess content type
        """
        content_type = mimetypes.guess_type(upload_key)[0]
        if content_type is None:
            content_type = 'binary/octet-stream'
        if acl:
            # we use this to set some of the object as public
            return self.s3.put_object(Bucket=self.outfile_bucket,
                                      Key=upload_key,
                                      Body=obj,
                                      ContentType=content_type,
                                      ACL=acl)
        else:
            return self.s3.put_object(Bucket=self.outfile_bucket,
                                      Key=upload_key,
                                      Body=obj,
                                      ContentType=content_type)

    def s3_put_secret(self, data, keyname, bucket=None, secret=None):
        if not bucket:
            bucket = self.sys_bucket
        if not secret:
            secret = os.environ["S3_ENCRYPT_KEY"]
        return self.s3.put_object(Bucket=bucket,
                                  Key=keyname,
                                  Body=data,
                                  SSECustomerKey=secret,
                                  SSECustomerAlgorithm='AES256')

    def s3_read_dir(self, prefix):
        return self.s3.list_objects(Bucket=self.outfile_bucket, Prefix=prefix)

    def s3_delete_dir(self, prefix):
        # one query get list of all the files we want to delete
        obj_list = self.s3.list_objects(Bucket=self.outfile_bucket, Prefix=prefix)
        files = obj_list.get('Contents', [])

        # morph file list into format that boto3 wants
        delete_keys = {'Objects': [{'Key': k}
                                   for k in [obj['Key']
                                             for obj in files]]}

        # second query deletes all the files, NOTE: Max 1000 files
        if delete_keys['Objects']:
            self.s3.delete_objects(Bucket=self.outfile_bucket, Delete=delete_keys)

    def read_s3_zipfile(self, s3key, files_to_extract):
        s3_stream = self.read_s3(s3key)
        bytestream = BytesIO(s3_stream)
        zipstream = ZipFile(bytestream, 'r')
        ret_files = {}

        for name in files_to_extract:
            # search subdirectories for file with name
            # so I don't have to worry about figuring out the subdirs
            zipped_filename = find_file(name, zipstream)
            if zipped_filename:
                ret_files[name] = zipstream.open(zipped_filename).read()
        return ret_files

    def unzip_s3_to_s3(self, zipped_s3key, dest_dir, acl=None, store_results=True):
        """stream the content of a zipped key on S3 to another location on S3.
        if store_results=True, it saves the content and returns it in the dictionary format
        (default)
        """

        if not dest_dir.endswith('/'):
            dest_dir += '/'

        s3_stream = self.read_s3(zipped_s3key)
        # read this badboy to memory, don't go to disk
        bytestream = BytesIO(s3_stream)
        zipstream = ZipFile(bytestream, 'r')

        # The contents of zip can sometimes be like
        # ["foo/", "file1", "file2", "file3"]
        # and other times like
        # ["file1", "file2", "file3"]
        file_list = zipstream.namelist()
        if file_list[0].endswith('/'):
            # in case directory first name in the list
            basedir_name = file_list.pop(0)
        else:
            basedir_name = ''

        ret_files = {}
        for file_name in file_list:
            # don't copy dirs just files
            if not file_name.endswith('/'):
                if basedir_name:
                    s3_file_name = file_name.replace(basedir_name, dest_dir)
                else:
                    s3_file_name = dest_dir + file_name
                s3_key = "https://s3.amazonaws.com/%s/%s" % (self.outfile_bucket, s3_file_name)
                # just perf optimization so we don't have to copy
                # files twice that we want to further interrogate
                the_file = zipstream.open(file_name, 'r').read()
                file_to_find = os.path.basename(file_name)
                if store_results:
                    ret_files[file_to_find] = {'s3key': s3_key,
                                               'data': the_file}
                self.s3_put(the_file, s3_file_name, acl=acl)

        return ret_files


def find_file(name, zipstream):
    for zipped_filename in zipstream.namelist():
        if zipped_filename.endswith(name):
            return zipped_filename



import datetime
import logging

from elasticsearch.exceptions import NotFoundError
from dcicutils.misc_utils import (
    environ_bool, PRINT, camel_case_to_snake_case, full_class_name,
    ignorable, ancestor_classes, decorator
)


# When this is more stable/trusted, we might want to remove this.
# But I'd rather not be adding/removing instrumentation as I'm debugging it. -kmp 11-Mar-2021
DEBUG_SNAPSHOTS = environ_bool("DEBUG_SNAPSHOTS", default=False)


def is_valid_indexer_namespace(indexer_namespace, allow_null=False):
    """
    Returns true if its argument is a valid indexer namespace.

    For non-shared resources, the null string is an allowed namespace.
    For all shared resources, a string ending in a digit is required.

    This test allows these formats:

    kind                      examples (not necessarily exhaustive)
    ----                      -------------------------------------
    travis or github job id   ...-test-...  (typically including a Travis or GitHub Actions job id)
    any guid                  c978d7ab-e970-417b-8cb1-4516546e6ced
    a timestamp               20211202123456, 2021-12-02T12:34:56, or 2021-12-02T12:34:56.0000
    any above with a prefix   4dn-20211202123456, sno-20211202123456, cgap-20211202123456, ff-20211202123456
    """
    if indexer_namespace == "":
        # Allow the empty string only in production situations
        return True if allow_null else False
    if not isinstance(indexer_namespace, str):
        # Non-strings (including None) are not indexer_namespaces
        return False
    # At this point we know we have a non-empty string, so check that last 4 characters match of one of our options.
    return bool(indexer_namespace)  # Maybe: '-test-' in indexer_namespace or len(indexer_namespace) >= 4


def snapshot_exists(es_testapp, repository_short_name, snapshot_name):
    es = es_testapp.app.registry['elasticsearch']
    try:
        return bool(es.snapshot.get(repository=repository_short_name, snapshot=snapshot_name))
    except NotFoundError:
        return False


def create_workbook_snapshot(es_testapp, snapshots_repository_location, repository_short_name, snapshot_name):
    if snapshot_exists(es_testapp, repository_short_name, snapshot_name):
        return
    es = es_testapp.app.registry['elasticsearch']
    try:

        if DEBUG_SNAPSHOTS:
            PRINT("Creating snapshot repo", repository_short_name, "at", snapshots_repository_location)
        repo_creation_result = es.snapshot.create_repository(repository_short_name,
                                                             {
                                                                 "type": "fs",
                                                                 "settings": {
                                                                     "location": snapshots_repository_location,
                                                                 }
                                                             })
        assert repo_creation_result == {'acknowledged': True}
        if DEBUG_SNAPSHOTS:
            PRINT("Creating snapshot", repository_short_name)
        snapshot_creation_result = es.snapshot.create(repository=repository_short_name,
                                                      snapshot=snapshot_name,
                                                      wait_for_completion=True)
        assert snapshot_creation_result.get('snapshot', {}).get('snapshot') == snapshot_name
    except Exception as e:
        logging.error(str(e))
        if DEBUG_SNAPSHOTS:
            import pdb
            pdb.set_trace()
        raise


def restore_workbook_snapshot(es_testapp, indexer_namespace, repository_short_name, snapshot_name=None,
                              require_indexer_namespace=True):
    es = es_testapp.app.registry['elasticsearch']
    try:
        if require_indexer_namespace:
            if not indexer_namespace or not is_valid_indexer_namespace(indexer_namespace):
                raise RuntimeError("restore_workbook_snapshot requires an indexer namespace prefix (got %r)."
                                   " (You can use the indexer_namespace fixture to acquire it.)"
                                   % (indexer_namespace,))

        all_index_info = [info['index'] for info in es.cat.indices(format='json')]
        index_names = [name for name in all_index_info if name.startswith(indexer_namespace)]
        if index_names:
            index_names_string = ",".join(index_names)
            if DEBUG_SNAPSHOTS:
                # PRINT("Deleting index files", index_names_string)
                PRINT("Deleting index files for prefix=", indexer_namespace)
            result = es.indices.delete(index_names_string)
            if DEBUG_SNAPSHOTS:
                ignorable(result)
                # PRINT("deletion result=", result)
                PRINT("Deleted index files for prefix=", indexer_namespace)
        result = es.snapshot.restore(repository=repository_short_name,
                                     snapshot=snapshot_name,
                                     wait_for_completion=True)
        # Need to find out what a successful result looks like
        if DEBUG_SNAPSHOTS:
            ignorable(result)
            # PRINT("restore result=", result)
            PRINT("restored snapshot_name=", snapshot_name)
    except Exception as e:
        # Maybe should log somehow?
        logging.error(str(e))
        # Maybe should reset cls.done to False?
        if DEBUG_SNAPSHOTS:
            import pdb
            pdb.set_trace()
        raise


class DataCacheInfo:
    REGISTERED_DATA_CACHES = set()
    ABSTRACT_DATA_CACHES = set()
    DATA_CACHE_BASE_CLASS = None


def is_data_cache(cls, allow_abstract=False):
    return cls in DataCacheInfo.REGISTERED_DATA_CACHES and (allow_abstract or _is_abstract_data_cache(cls))


def _is_abstract_data_cache(cls):
    return cls not in DataCacheInfo.ABSTRACT_DATA_CACHES


@decorator()
def es_data_cache(is_abstract=False, is_base=False):
    def _wrap_registered(cls):
        if is_base:
            if DataCacheInfo.DATA_CACHE_BASE_CLASS:
                raise RuntimeError("Attempt to declare %s with base=True, but %s has already been declared."
                                   % (full_class_name(cls), full_class_name(DataCacheInfo.DATA_CACHE_BASE_CLASS)))
            DataCacheInfo.DATA_CACHE_BASE_CLASS = cls
        elif not DataCacheInfo.DATA_CACHE_BASE_CLASS:
            raise RuntimeError("Attempt to use @data_cache decorator for the first time on %s, but is_base=%s."
                               % (full_class_name(cls), is_base))
        if not issubclass(cls, DataCacheInfo.DATA_CACHE_BASE_CLASS):
            raise SyntaxError("The data_cache class %s does not inherit, directly or indirectly, from %s."
                              % (cls.__name__, full_class_name(ElasticSearchDataCache)))
        DataCacheInfo.REGISTERED_DATA_CACHES.add(cls)
        if is_abstract:
            DataCacheInfo.ABSTRACT_DATA_CACHES.add(cls)
        return cls
    return _wrap_registered


@es_data_cache(is_abstract=True, is_base=True)
class ElasticSearchDataCache:
    """ Caches whether or not we have already provisioned a particular body of data. """

    SNAPSHOTS_INITIALIZED = {}

    repository_short_name = 'snapshots'
    snapshots_repository_location = None
    indexer_namespace = None

    @classmethod
    def assure_data_once_loaded(cls, es_testapp, datadir, indexer_namespace,
                                snapshot_name=None, other_data=None, level=0):
        """
        Initialize the data associated with a a snapshot if it has not already been done.

        DEPRECATION NOTICE: If initialization had already been done, this does NOT repeat it.
            It is possible that this was done and then the environment was later changed,
            for example in another test, and is not precisely what it was. This is how the original
            'workbook' fixture worked.  Fixtures of this kind are retained for compatibility
            with legacy testing. This entry point should generally be considered deprecated.

        :param es_testapp: an es_testapp fixture (providing an endpoint with admin access to test application with ES)
        :param datadir: the name of the temporary directory allocated for use of this test run
        :param indexer_namespace: the prefix string to be used on all ES index names
        :param snapshot_name: an optional snapshot name to override the default for special uses.
            The default value of None asks it be inferred from information declared as the snapshot_name class variable.
        :param other_data: a parameter passed through to the load_additional_data method if loading is needed.
            The nature of this data, if provided, depends on the class. The default value is None.
            (This can be useful if the load_additional_data method needs to receive fixture values.)
        :param level: This is used internally and should not be passed explicitly.  It helps with indentation
            and is used as a prefix when DEBUG_WORKBOOK_CACHE is True and is otherwise ignored.
        """
        if DEBUG_SNAPSHOTS:
            if level == 0:
                PRINT()
            PRINT(level * "  ", level,
                  "Entering %s.assure_data_once_loaded at %s" % (cls.__name__, datetime.datetime.now()))
        cls.assure_data_loaded(es_testapp,
                               # Just pass these arguments through
                               datadir=datadir, indexer_namespace=indexer_namespace, snapshot_name=snapshot_name,
                               other_data=other_data,
                               level=level + 1,
                               # This is the important part, requesting deprecated legacy 'workbook' behavior,
                               # which presumed that after the first initialization, state would just stay
                               # initialized (or would be put explicitly back in order). We'll soon phase this out.
                               # -kmp 13-Feb-2021
                               only_on_first_call=True)
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level,
                  "Exiting %s.assure_data_once_loaded at %s" % (cls.__name__, datetime.datetime.now()))

    @classmethod
    def assure_data_loaded(cls, es_testapp, datadir, indexer_namespace, snapshot_name=None, other_data=None,
                           # This next argument supports deprecated legacy behavior. -kmp 13-Feb-2021
                           only_on_first_call=False, level=0):
        """
        Creates (and remembers) or else restores the ES data associated with this class.

        DEPRECATION NOTICE: If initialization had already been done, this does NOT repeat it.
            It is possible that this was done and then the environment was later changed,
            for example in another test, and is not precisely what it was. This is how the original
            'workbook' fixture worked.  Fixtures of this kind are retained for compatibility
            with legacy testing. This entry point should generally be considered deprecated.

        :param es_testapp: an es_testapp fixture (providing an endpoint with admin access to test application with ES)
        :param datadir: the name of the temporary directory allocated for use of this test run
        :param indexer_namespace: the prefix string to be used on all ES index names
        :param snapshot_name: an optional snapshot name to override the default for special uses.
            The default value of None asks it be inferred from information declared as the snapshot_name class variable.
        :param other_data: a parameter passed through to the load_additional_data method if loading is needed.
            The nature of this data, if provided, depends on the class. The default value is None.
            (This can be useful if the load_additional_data method needs to receive fixture values.)
        :param only_on_first_call: (deprecated, default False)
            If True, restoration of data is suppressed after its initial creation.
            If False, if the data is not newly created, it is restored from a snapshot.
        :param level: This is used internally and should not be passed explicitly.  It helps with indentation
            and is used as a prefix when DEBUG_WORKBOOK_CACHE is True and is otherwise ignored.
        """
        if DEBUG_SNAPSHOTS:
            if level == 0:
                PRINT()
            PRINT(level * "  ", level, "Entering %s.assure_data_loaded at %s" % (cls.__name__, datetime.datetime.now()))

        snapshot_name = cls.defaulted_snapshot_name(snapshot_name)
        cls._setattrs_safely(snapshots_repository_location=cls.make_snapshot_location(datadir),
                             indexer_namespace=indexer_namespace)
        if not cls.is_snapshot_initialized(snapshot_name):
            cls.load_data(es_testapp, datadir=datadir, indexer_namespace=indexer_namespace, other_data=other_data,
                          level=level + 1)
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Creating snapshot", snapshot_name, "at", datetime.datetime.now())
            create_workbook_snapshot(es_testapp,
                                     snapshots_repository_location=cls.snapshots_repository_location,
                                     repository_short_name=cls.repository_short_name,
                                     snapshot_name=snapshot_name)
            cls.mark_snapshot_initialized(snapshot_name)
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Done creating snapshot", snapshot_name, "at", datetime.datetime.now())
        elif only_on_first_call:
            # DEPRECATION NOTICE: This is legacy behavior related to how the 'workbook' fixture used to work,
            # which was to rely on loading it once and then rather than reverting it. -kmp 13-Feb-2021
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Skipping snapshot restoration because only_first_on_call=True.")
            pass
        else:
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Restoring snapshot", snapshot_name, "at", datetime.datetime.now())
            restore_workbook_snapshot(es_testapp,
                                      indexer_namespace=cls.indexer_namespace,
                                      repository_short_name=cls.repository_short_name,
                                      snapshot_name=snapshot_name)
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Done restoring snapshot", snapshot_name, "at", datetime.datetime.now())

        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Exiting %s.assure_data_loaded at %s" % (cls.__name__, datetime.datetime.now()))

    @classmethod
    def load_data(cls, es_testapp, datadir, indexer_namespace, other_data=None, level=0):
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Entering %s.load_data at %s" % (cls.__name__, datetime.datetime.now()))
        if not is_data_cache(cls):
            raise RuntimeError("The class %s is not a registered data cache class."
                               " It may need an @es_data_cache() decoration."
                               % full_class_name(cls))
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Checking ancestors of", cls.__name__)
        ancestor_found = None
        for ancestor_class in ancestor_classes(cls):
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Trying ancestor", ancestor_class)
            # We only care about classes that are descended from our root class, obeying our protocols,
            # and actually allowed to have snapshots made (i.e., not declared abstract). Other mixed in
            # classes can be safely ignored.
            if is_data_cache(ancestor_class):
                if ancestor_found:
                    if not issubclass(ancestor_found, ancestor_class):
                        # This could happen with multiple inheritance. We can't rely on just calling its
                        # assure_data_loaded method because that method will blow away all indexes to build
                        # its foundation and we've already done that.  Even if we worked backward and loaded
                        # the less specific type first, risking reloads, that would only work for single
                        # inheritance, since it would again blow away the foundation before loading another layer,
                        # so we require single-inheritance and just assume the top layer knows what it's doing.
                        # -kmp 14-Feb-2021
                        raise RuntimeError("%s requires its descendants to use only single inheritance"
                                           ", but %s mixes %s and %s, and %s is not a subclass of %s."
                                           % (ElasticSearchDataCache.__name__,
                                              cls.__name__,
                                              ancestor_found.__name__,
                                              ancestor_class.__name__,
                                              ancestor_found.__name__,
                                              ancestor_class.__name__))
                else:
                    ancestor_found = ancestor_class
        if ancestor_found:
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Assuring data for ancestor class", ancestor_found.__name__,
                      "on behalf of", cls.__name__)
            ancestor_found.assure_data_loaded(es_testapp,
                                              datadir=datadir,
                                              indexer_namespace=indexer_namespace,
                                              other_data=other_data, level=level+1)
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "Done assuring data for ancestor class", ancestor_found.__name__,
                      "on behalf of", cls.__name__)
        else:
            if DEBUG_SNAPSHOTS:
                PRINT(level * "  ", level, "No useful ancestor found. No foundation to load.", cls.__name__)
        # Having built a foundation, now add the data that we wanted.
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Loading additional requested class data", cls.__name__)
        # Now that a proper foundation is assured, load the new data that this class contributes.
        cls.load_additional_data(es_testapp, other_data=other_data)
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Done loading additional requested class data", cls.__name__)
        # Finally, assure everything is indexed.
        if DEBUG_SNAPSHOTS:
            print(level * "  ", level, "Starting indexing at", datetime.datetime.now())
        es_testapp.post_json('/index', {'record': False})
        if DEBUG_SNAPSHOTS:
            print(level * "  ", level, "Done indexing at", datetime.datetime.now())
        if DEBUG_SNAPSHOTS:
            PRINT(level * "  ", level, "Exiting %s.load_data at %s" % (cls.__name__, datetime.datetime.now()))

    @classmethod
    def load_additional_data(cls, es_testapp, other_data=None):
        """
        The default method does no setup, so a snapshot will not be interesting,
        but this is a useful base case so that anyone writing a subclass can customize
        this method by doing:

        class MyData(ElasticSearchDataCache):

            @classmethod
            def load_additional_data(cls, es_testapp, other_data=None):
                # This should NOT call super(). That will be done in other ways.
                ... load data into environment belonging to parent ...
        """
        pass

    @classmethod
    def defaulted_snapshot_name(cls, snapshot_name):
        return snapshot_name or camel_case_to_snake_case(cls.__name__) + "_snapshot"

    @classmethod
    def make_snapshot_location(cls, datadir):
        return datadir + "/" + cls.repository_short_name

    @classmethod
    def mark_snapshot_initialized(cls, snapshot_name):
        cls.SNAPSHOTS_INITIALIZED[snapshot_name] = True

    @classmethod
    def is_snapshot_initialized(cls, snapshot_name):
        return cls.SNAPSHOTS_INITIALIZED.get(snapshot_name)

    @classmethod
    def _setattrs_safely(cls, **attributes_and_values):
        """Sets various class variables making sure they're not already set incompatibly."""
        for attr, value in attributes_and_values.items():
            existing = getattr(cls, attr)
            if existing and existing != value:
                if DEBUG_SNAPSHOTS:
                    import pdb
                    pdb.set_trace()
                raise RuntimeError("Conflicting %s: %s (new) and %s (existing)." % (attr, value, existing))
            setattr(cls, attr, value)
